/*
** MFCHESS 1.0 (C) 1998 Mats Forsn
**
** AI
*/

#include "ai.hpp"
#include <stdlib.h>

int AI::accept_draw( int playercolor )
{
	return 1; // Check for close mate later
}

void AI::reset()
{
   int x;

   for( x = 0; x < 0xFFFF; x++ )
   	historydata[ x ] = 0;

   nodes_visited = 0;

   best_move.reset();
}

int AI::move_score( Move m )
{
	int taken;

   if( (taken = m.what_piece_taken()) != 0 ) // Capture move
   	return ((taken*10) - m.what_piece_moved()) << 5;

   if( m.is_promotion() ) // Promotion move
   	return (1 << 5);

   // Historydata
   return historydata[ m.where() ];
}

void AI::sort_best_move( int from, MoveList &moves )
{
	int best = -1, bestnum = -1, score, x = from;

   if( from == 0 )
   	return;

   do
   {
   	score = move_score( moves.m[x] );
   	if( score > best )
      {
      	best = score;
         bestnum = x;
      }
   }
   while( --x >= 0 );

   if( bestnum == from )
   	return;

   // Swap da moves
   Move temp = moves.m[from];
   moves.m[from] = moves.m[bestnum];
   moves.m[bestnum] = temp;
}

int AI::negamax( int playercolor, int depth, int alpha, int beta )
{
	static MoveList moves[ MAX_DEPTH ];
	int result, curmove, sort = 5;

   nodes_visited++;

   if( depth == 0 )
   	return board->Evaluate(playercolor);

   moves[depth].num = 0;
   board->GenerateMoves( playercolor, moves[depth] );

	curmove = moves[depth].num;

	while( curmove-- )
   {
   	if( sort ) // Only sort the first 8 moves
      {
   		sort_best_move( curmove, moves[depth] ); // Get the best move
         sort--;
      }

   	board->ApplyMove( moves[depth].m[curmove] );

      if( board->InCheck( playercolor ) )
      {
      	board->UndoMove( moves[depth].m[curmove] );
			moves[depth].num--;
 			continue; // No check moves please
      }

		// Recurse using nega-max variant of alpha-beta
      result = - negamax( SWAP(playercolor), depth-1, -beta, -alpha );

		board->UndoMove( moves[depth].m[curmove] );

      if( result > alpha )
      {
      	alpha = result;
         if( depth == maxdepth )
         	best_move = moves[depth].m[curmove];
      }

      if( alpha >= beta )
      {
         historydata[ moves[depth].m[curmove].where() ] +=
         														( 1 << (maxdepth - depth) );
         return beta;
      }

   }

   if( moves[depth].num == 0 ) // Mate or Draw
   {
   	if( board->InCheck(playercolor) )
   		return -INF + 100; // Very bad.. Mate
      else
      	return 0; // draw
   }

	return alpha;
}

Move AI::get_computer_move( int playercolor )
{
	clock_t start, stop;

	reset();

   board->InitEval();

   if( usebook /* && trybook */ ) // Test: Always check book
   {
   	start = clock();
   	trybook = book.GetBookMove( best_move, board, playercolor );
      stop = clock();
      if( trybook )
      { // Success
      	time = stop - start;
         nodes_visited = -1;
      	return best_move;
      }
   }

   start = clock();
   negamax( playercolor, maxdepth, -INF, +INF );
   stop = clock();

   // Calc time
   time = stop - start;

   return best_move;
}

/* EOF */

